@@ -22,6 +22,7 @@
 #include <config.h>
 #endif
 
+#include <stdio.h>
 #include <ibus.h>
 #include <gio/gio.h>
 #include <hangul.h>
@@ -44,18 +45,49 @@ enum {
     INPUT_MODE_COUNT,
 };
 
+/**
+ * ibus-hangul supports 3 preedit modes.
+ */
+typedef enum {
+    /**
+     * @brief zero length preedit mode
+     * An option to use zero length preedit text.
+     * ibus-hangul will utilize surrounding text feature to draw "composing text".
+     * So the "composing text" will not be shown in preedit style but normal text.
+     */
+    PREEDIT_MODE_NONE,
+    /**
+     * @brief syllable preedit mode
+     * An option to use syllable length preedit text.
+     * This option utilizes preedit text feature.
+     */
+    PREEDIT_MODE_SYLLABLE,
+    /**
+     * @brief word preedit mode
+     * An option to use word length preedit text.
+     */
+    PREEDIT_MODE_WORD,
+} IBusHangulPreeditMode;
+
 struct _IBusHangulEngine {
     IBusEngineSimple parent;
 
     /* members */
+    /* unique context id */
+    guint id;
+
     HangulInputContext *context;
     UString* preedit;
+    /* Every instance's preedit_mode may be different from global settings.
+     * So we need a value for each instance. */
+    IBusHangulPreeditMode preedit_mode;
     int input_mode;
     unsigned int input_purpose;
     gboolean hanja_mode;
     HanjaList* hanja_list;
     int last_lookup_method;
 
+    guint caps;
     IBusLookupTable *table;
 
     IBusProperty    *prop_hangul_mode;
@@ -112,10 +144,10 @@ static void ibus_engine_set_cursor_location (IBusEngine             *engine,
                                              gint                    y,
                                              gint                    w,
                                              gint                    h);
+#endif
 static void ibus_hangul_engine_set_capabilities
                                             (IBusEngine             *engine,
                                              guint                   caps);
-#endif
 static void ibus_hangul_engine_page_up      (IBusEngine             *engine);
 static void ibus_hangul_engine_page_down    (IBusEngine             *engine);
 static void ibus_hangul_engine_cursor_up    (IBusEngine             *engine);
@@ -148,6 +180,8 @@ static void ibus_hangul_engine_clear_preedit_text
                                             (IBusHangulEngine       *hangul);
 static void ibus_hangul_engine_update_preedit_text
                                             (IBusHangulEngine       *hangul);
+static void ibus_hangul_engine_process_commit_and_edit
+                                            (IBusHangulEngine       *hangul);
 
 static void ibus_hangul_engine_update_lookup_table
                                             (IBusHangulEngine       *hangul);
@@ -197,7 +231,10 @@ static gboolean hotkey_list_has_modifier    (HotkeyList           *list,
 
 static glong ucschar_strlen (const ucschar* str);
 
+static gint ibus_version[3] = { IBUS_MAJOR_VERSION, IBUS_MINOR_VERSION, IBUS_MICRO_VERSION };
+
 static IBusEngineSimpleClass *parent_class = NULL;
+static guint last_context_id = 0;
 static HanjaTable *hanja_table = NULL;
 static HanjaTable *symbol_table = NULL;
 static GSettings *settings_hangul = NULL;
@@ -215,8 +252,22 @@ static gboolean disable_latin_mode = FALSE;
 static int initial_input_mode = INPUT_MODE_LATIN;
 /**
  * whether to use event forwarding workaround
+ * See: https://github.com/libhangul/ibus-hangul/issues/42
  */
 static gboolean use_event_forwarding = TRUE;
+/**
+ * whether to use client commit
+ * See: https://github.com/libhangul/ibus-hangul/pull/68
+ */
+static gboolean use_client_commit = FALSE;
+
+/**
+ * global preedit mode
+ * This option may have a value, one of the IBusHangulPreeditMode.
+ * See: https://github.com/libhangul/ibus-hangul/issues/69
+ */
+static IBusHangulPreeditMode global_preedit_mode = PREEDIT_MODE_SYLLABLE;
+
 
 static glong
 ucschar_strlen (const ucschar* str)
@@ -227,6 +278,86 @@ ucschar_strlen (const ucschar* str)
     return p - str;
 }
 
+static void
+check_ibus_version ()
+{
+    gboolean retval;
+    gchar* standard_output = NULL;
+    const gchar* version_str;
+    gchar** version_str_array;
+    gint version[3];
+
+    retval = g_spawn_command_line_sync ("ibus version", &standard_output,
+                                        NULL, NULL, NULL);
+    if (!retval)
+        goto fail;
+
+    version_str = strpbrk (standard_output, " \t");
+    if (version_str == NULL) {
+        g_free (standard_output);
+        goto fail;
+    }
+
+    while (*version_str == ' ' || *version_str == '\t')
+        version_str++;
+
+    version_str_array = g_strsplit (version_str, ".", 3);
+    g_free (standard_output);
+    if (version_str_array == NULL) {
+        goto fail;
+    }
+
+    version[0] = 0;
+    version[1] = 0;
+    version[2] = 0;
+    for (gint i = 0; version_str_array[i] != NULL; ++i) {
+        version[i] = g_ascii_strtoll (version_str_array[i], NULL, 10);
+    }
+    g_strfreev (version_str_array);
+
+    if (version[0] == 0 && version[1] == 0 && version[2] == 0) {
+        goto fail;
+    }
+
+    ibus_version[0] = version[0];
+    ibus_version[1] = version[1];
+    ibus_version[2] = version[2];
+    g_debug ("ibus version detected: %d.%d.%d",
+            ibus_version[0], ibus_version[1], ibus_version[2]);
+    return;
+
+fail:
+    g_debug ("ibus version detection failed: use default value: %d.%d.%d",
+            ibus_version[0], ibus_version[1], ibus_version[2]);
+}
+
+static gboolean
+ibus_hangul_check_ibus_version (gint required_major,
+                                gint required_minor,
+                                gint required_micro)
+{
+    gint major = ibus_version[0];
+    gint minor = ibus_version[1];
+    gint micro = ibus_version[2];
+
+    return major > required_major ||
+        (major == required_major && minor > required_minor) ||
+        (major == required_major && minor == required_minor &&
+         micro >= required_micro);
+}
+
+static gboolean
+check_client_commit ()
+{
+    gboolean client_commit = FALSE;
+
+    client_commit = ibus_hangul_check_ibus_version (1, 5, 20);
+
+    g_debug ("client_commit: %d", client_commit);
+
+    return client_commit;
+}
+
 GType
 ibus_hangul_engine_get_type (void)
 {
@@ -259,10 +390,14 @@ ibus_hangul_init (IBusBus *bus)
 {
     GVariant* value = NULL;
 
+    last_context_id = 0;
+
     hanja_table = hanja_table_load (NULL);
 
     symbol_table = hanja_table_load (IBUSHANGUL_DATADIR "/data/symbol.txt");
 
+    check_ibus_version ();
+
     settings_hangul = g_settings_new ("org.freedesktop.ibus.engine.hangul");
     settings_panel = g_settings_new ("org.freedesktop.ibus.panel");
 
@@ -349,6 +484,19 @@ ibus_hangul_init (IBusBus *bus)
         g_clear_pointer (&value, g_variant_unref);
     }
 
+    value = g_settings_get_value (settings_hangul, "preedit-mode");
+    if (value != NULL) {
+        const gchar* str = g_variant_get_string (value, NULL);
+        if (strcmp(str, "none") == 0) {
+            global_preedit_mode = PREEDIT_MODE_NONE;
+        } else if (strcmp(str, "word") == 0) {
+            global_preedit_mode = PREEDIT_MODE_WORD;
+        } else {
+            global_preedit_mode = PREEDIT_MODE_SYLLABLE;
+        }
+        g_clear_pointer (&value, g_variant_unref);
+    }
+
     value = g_settings_get_value (settings_panel, "lookup-table-orientation");
     if (value != NULL) {
         lookup_table_orientation = g_variant_get_int32(value);
@@ -356,11 +504,16 @@ ibus_hangul_init (IBusBus *bus)
     }
 
     keymap = ibus_keymap_get("us");
+    use_client_commit = check_client_commit ();
+
+    g_debug ("init");
 }
 
 void
 ibus_hangul_exit (void)
 {
+    g_debug ("exit");
+
     if (keymap != NULL) {
 	g_object_unref(keymap);
 	keymap = NULL;
@@ -384,6 +537,19 @@ ibus_hangul_exit (void)
     hangul_keyboard = NULL;
 }
 
+/*
+static void
+ibus_hangul_engine_set_surrounding_text (IBusEngine     *engine,
+                                         IBusText       *text,
+                                         guint           cursor_index,
+                                         guint           anchor_pos)
+
+{
+    g_debug ("set_surrounding_text: %s %d %d",
+            ibus_text_get_text (text), cursor_index, anchor_pos);
+}
+*/
+
 static void
 ibus_hangul_engine_class_init (IBusHangulEngineClass *klass)
 {
@@ -405,6 +571,7 @@ ibus_hangul_engine_class_init (IBusHangulEngineClass *klass)
     engine_class->focus_in = ibus_hangul_engine_focus_in;
     engine_class->focus_out = ibus_hangul_engine_focus_out;
 
+    engine_class->set_capabilities = ibus_hangul_engine_set_capabilities;
     engine_class->page_up = ibus_hangul_engine_page_up;
     engine_class->page_down = ibus_hangul_engine_page_down;
 
@@ -414,6 +581,7 @@ ibus_hangul_engine_class_init (IBusHangulEngineClass *klass)
     engine_class->property_activate = ibus_hangul_engine_property_activate;
 
     engine_class->candidate_clicked = ibus_hangul_engine_candidate_clicked;
+    //engine_class->set_surrounding_text = ibus_hangul_engine_set_surrounding_text;
     engine_class->set_content_type = ibus_hangul_engine_set_content_type;
 }
 
@@ -425,16 +593,21 @@ ibus_hangul_engine_init (IBusHangulEngine *hangul)
     IBusText* tooltip;
     IBusText* symbol;
 
+    hangul->id = last_context_id;
+    ++last_context_id;
+
     hangul->context = hangul_ic_new (hangul_keyboard->str);
     hangul_ic_connect_callback (hangul->context, "transition",
                                 ibus_hangul_engine_on_transition, hangul);
 
     hangul->preedit = ustring_new();
+    hangul->preedit_mode = global_preedit_mode;
     hangul->hanja_list = NULL;
     hangul->input_mode = initial_input_mode;
     hangul->input_purpose = IBUS_INPUT_PURPOSE_FREE_FORM;
     hangul->hanja_mode = FALSE;
     hangul->last_lookup_method = LOOKUP_METHOD_PREFIX;
+    hangul->caps = 0;
 
     if (disable_latin_mode) {
         hangul->input_mode = INPUT_MODE_HANGUL;
@@ -487,6 +660,8 @@ ibus_hangul_engine_init (IBusHangulEngine *hangul)
                       G_CALLBACK (settings_changed), hangul);
     g_signal_connect (settings_panel, "changed",
                       G_CALLBACK (settings_changed), hangul);
+
+    g_debug ("context new:%u", hangul->id);
 }
 
 static GObject*
@@ -510,6 +685,8 @@ ibus_hangul_engine_destroy (IBusHangulEngine *hangul)
     int i;
     IBusText **symbols;
 
+    g_debug ("context delete:%u", hangul->id);
+
     if (hangul->prop_hangul_mode) {
         g_object_unref (hangul->prop_hangul_mode);
         hangul->prop_hangul_mode = NULL;
@@ -525,6 +702,11 @@ ibus_hangul_engine_destroy (IBusHangulEngine *hangul)
         hangul->prop_list = NULL;
     }
 
+    if (hangul->preedit != NULL) {
+        ustring_delete (hangul->preedit);
+        hangul->preedit = NULL;
+    }
+
     if (hangul->table) {
         g_object_unref (hangul->table);
         hangul->table = NULL;
@@ -546,6 +728,70 @@ ibus_hangul_engine_destroy (IBusHangulEngine *hangul)
     IBUS_OBJECT_CLASS (parent_class)->destroy ((IBusObject *)hangul);
 }
 
+/**
+ * @brief a function to check whether the caret has moved
+ *
+ * An ibus client should inform to the engine that the caret has moved
+ * by calling reset method. But many implementations don't follow this rule.
+ * So we need to check whether the caret pos is changed.
+ * If so, we will reset the context.
+ *
+ * Usually we don't need this function. Only when the preedit mode is
+ * PREEDIT_MODE_NONE, it is critical to have same surrounding text and
+ * internal cached preedit text.
+ */
+static void
+ibus_hangul_engine_check_caret_pos_sanity (IBusHangulEngine *hangul)
+{
+    size_t preedit_len = ustring_length (hangul->preedit);
+    if (preedit_len == 0)
+        return;
+
+    IBusText* ibus_text = NULL;
+    guint cursor_pos = 0;
+    guint anchor_pos = 0;
+    ibus_engine_get_surrounding_text ((IBusEngine *)hangul,
+            &ibus_text, &cursor_pos, &anchor_pos);
+
+    const gchar* text = ibus_text_get_text (ibus_text);
+    if (text == NULL || cursor_pos == 0)
+        return;
+
+    const gchar* text_on_cursor = g_utf8_offset_to_pointer (text, cursor_pos - 1);
+    gchar* preedit_utf8 = ustring_to_utf8 (hangul->preedit, -1);
+    if (preedit_utf8 == NULL) {
+        g_object_unref (ibus_text);
+        return;
+    }
+
+    size_t n = strlen(preedit_utf8);
+    // Ok, Just comparing text value is not perfect. But any other idea?
+    if (strncmp(preedit_utf8, text_on_cursor, n) != 0) {
+        // If the text_on_cursor is different from preedit cache, there's a possibility
+        // that the cursor was moved by the user. Then we need to reset the context.
+        hangul_ic_reset (hangul->context);
+        ustring_clear (hangul->preedit);
+    }
+    g_free (preedit_utf8);
+
+    g_object_unref (ibus_text);
+}
+
+static void
+ibus_hangul_engine_update_preedit_mode (IBusHangulEngine *hangul)
+{
+    if (global_preedit_mode == PREEDIT_MODE_NONE &&
+        !(hangul->caps & IBUS_CAP_SURROUNDING_TEXT)) {
+        // If an instance doesn't support surrounding text, ibus-hangul will change
+        // preedit mode of this instance to PREEDIT_MODE_SYLLABLE.
+        // Because without surrounding text feature, users cannot see "composing text".
+        // This is pretty inconvenient for korean users.
+        hangul->preedit_mode = PREEDIT_MODE_SYLLABLE;
+    } else {
+        hangul->preedit_mode = global_preedit_mode;
+    }
+}
+
 static void
 ibus_hangul_engine_clear_preedit_text (IBusHangulEngine *hangul)
 {
@@ -563,6 +809,10 @@ ibus_hangul_engine_update_preedit_text (IBusHangulEngine *hangul)
     UString *preedit;
     gint preedit_len;
 
+    if (hangul->preedit_mode == PREEDIT_MODE_NONE) {
+        return;
+    }
+
     // ibus-hangul's preedit string is made up of ibus context's
     // internal preedit string and libhangul's preedit string.
     // libhangul only supports one syllable preedit string.
@@ -603,6 +853,77 @@ ibus_hangul_engine_update_preedit_text (IBusHangulEngine *hangul)
     ustring_delete(preedit);
 }
 
+static void
+ibus_hangul_engine_process_commit_and_edit (IBusHangulEngine *hangul)
+{
+    // commit current commit_text + preedit_text
+    const ucschar *hic_commit_text = hangul_ic_get_commit_string (hangul->context);
+    const ucschar *hic_preedit_text = hangul_ic_get_preedit_string (hangul->context);
+
+    UString *commit_text = ustring_new ();
+    ustring_append_ucs4 (commit_text, hic_commit_text, -1);
+    ustring_append_ucs4 (commit_text, hic_preedit_text, -1);
+
+    // commit only when the final result is different from preedit text cache
+    if (ustring_compare (commit_text, hangul->preedit) != 0) {
+        IBusEngine *engine = (IBusEngine *)hangul;
+
+        // remove composing text
+        guint preedit_text_len = ustring_length (hangul->preedit);
+        ibus_engine_delete_surrounding_text (engine, -preedit_text_len, preedit_text_len);
+
+        const ucschar *s = ustring_begin (commit_text);
+        IBusText *text = ibus_text_new_from_ucs4 (s);
+        ibus_engine_commit_text (engine, text);
+    }
+
+    ustring_delete (commit_text);
+
+    // update preedit_text cache
+    ustring_clear (hangul->preedit);
+    ustring_append_ucs4 (hangul->preedit, hic_preedit_text, -1);
+}
+
+static void
+ibus_hangul_engine_process_edit_and_commit (IBusHangulEngine *hangul)
+{
+    IBusEngine *engine = (IBusEngine *)hangul;
+
+    const ucschar *hic_commit_text = hangul_ic_get_commit_string (hangul->context);
+    const ucschar *hic_preedit_text = hangul_ic_get_preedit_string (hangul->context);
+
+    if (hangul->preedit_mode == PREEDIT_MODE_WORD || hangul->hanja_mode) {
+        ustring_append_ucs4 (hangul->preedit, hic_commit_text, -1);
+
+        if (hic_preedit_text == NULL || hic_preedit_text[0] == 0) {
+            if (ustring_length (hangul->preedit) > 0) {
+                IBusText *text;
+                const ucschar* preedit_text;
+
+                /* clear preedit text before commit */
+                ibus_hangul_engine_clear_preedit_text (hangul);
+
+                preedit_text = ustring_begin (hangul->preedit);
+                text = ibus_text_new_from_ucs4 ((gunichar*)preedit_text);
+                ibus_engine_commit_text (engine, text);
+                ustring_clear (hangul->preedit);
+            }
+        }
+    } else {
+        if (hic_commit_text != NULL && hic_commit_text[0] != 0) {
+            IBusText *text;
+
+            /* clear preedit text before commit */
+            ibus_hangul_engine_clear_preedit_text (hangul);
+
+            text = ibus_text_new_from_ucs4 (hic_commit_text);
+            ibus_engine_commit_text (engine, text);
+        }
+    }
+
+    ibus_hangul_engine_update_preedit_text (hangul);
+}
+
 static void
 ibus_hangul_engine_update_lookup_table_ui (IBusHangulEngine *hangul)
 {
@@ -668,7 +989,17 @@ ibus_hangul_engine_commit_current_candidate (IBusHangulEngine *hangul)
         /* remove hic preedit text */
         if (hic_preedit_len > 0) {
             hangul_ic_reset (hangul->context);
-            key_len -= hic_preedit_len;
+            if (hangul->preedit_mode == PREEDIT_MODE_NONE) {
+                if (preedit_len > hic_preedit_len) {
+                    guint pos = preedit_len - hic_preedit_len;
+                    ustring_erase (hangul->preedit, pos, hic_preedit_len);
+                } else {
+                    ustring_clear (hangul->preedit);
+                }
+                preedit_len = ustring_length(hangul->preedit);
+            } else {
+                key_len -= hic_preedit_len;
+            }
         }
 
         /* remove ibus preedit text */
@@ -763,6 +1094,7 @@ ibus_hangul_engine_lookup_hanja_table (const char* key, int method)
         break;
     }
 
+    g_debug("lookup hanja table: %s", key);
     return list;
 }
 
@@ -772,7 +1104,7 @@ ibus_hangul_engine_update_hanja_list (IBusHangulEngine *hangul)
     gchar* hanja_key;
     gchar* preedit_utf8;
     const ucschar* hic_preedit;
-    UString* preedit;
+    UString* preedit = NULL;
     int lookup_method;
     IBusText* ibus_text = NULL;
     guint cursor_pos = 0;
@@ -788,12 +1120,14 @@ ibus_hangul_engine_update_hanja_list (IBusHangulEngine *hangul)
     hanja_key = NULL;
     lookup_method = LOOKUP_METHOD_PREFIX;
 
-    preedit = ustring_dup (hangul->preedit);
-    ustring_append_ucs4 (preedit, hic_preedit, -1);
+    if (hangul->preedit_mode != PREEDIT_MODE_NONE) {
+        preedit = ustring_dup (hangul->preedit);
+        ustring_append_ucs4 (preedit, hic_preedit, -1);
+    }
 
-    if (ustring_length(preedit) > 0) {
+    if (preedit != NULL && ustring_length(preedit) > 0) {
         preedit_utf8 = ustring_to_utf8 (preedit, -1);
-        if (word_commit || hangul->hanja_mode) {
+        if (hangul->preedit_mode == PREEDIT_MODE_WORD || hangul->hanja_mode) {
             hanja_key = preedit_utf8;
             lookup_method = LOOKUP_METHOD_PREFIX;
         } else {
@@ -802,7 +1136,7 @@ ibus_hangul_engine_update_hanja_list (IBusHangulEngine *hangul)
                     &cursor_pos, &anchor_pos);
 
             substr = h_ibus_text_get_substring (ibus_text,
-                    cursor_pos - 64, cursor_pos);
+                    (glong)cursor_pos - 32, cursor_pos);
 
             if (substr != NULL) {
                 hanja_key = g_strconcat (substr, preedit_utf8, NULL);
@@ -822,7 +1156,7 @@ ibus_hangul_engine_update_hanja_list (IBusHangulEngine *hangul)
             lookup_method = LOOKUP_METHOD_EXACT;
         } else {
             hanja_key = h_ibus_text_get_substring (ibus_text,
-                    cursor_pos - 64, cursor_pos);
+                    (glong)cursor_pos - 32, cursor_pos);
             lookup_method = LOOKUP_METHOD_SUFFIX;
         }
     }
@@ -834,7 +1168,8 @@ ibus_hangul_engine_update_hanja_list (IBusHangulEngine *hangul)
         g_free (hanja_key);
     }
 
-    ustring_delete (preedit);
+    if (preedit != NULL)
+        ustring_delete (preedit);
 
     if (ibus_text != NULL)
         g_object_unref (ibus_text);
@@ -1045,7 +1380,7 @@ ibus_hangul_engine_process_key_event (IBusEngine     *engine,
 
     guint mask;
     gboolean retval;
-    const ucschar *str;
+    guint orig_keyval = keyval;
 
     if (modifiers & IBUS_RELEASE_MASK)
         return FALSE;
@@ -1062,6 +1397,21 @@ ibus_hangul_engine_process_key_event (IBusEngine     *engine,
     if (hangul->input_purpose == IBUS_INPUT_PURPOSE_PASSWORD)
         return IBUS_ENGINE_CLASS (parent_class)->process_key_event (engine, keyval, keycode, modifiers);
 
+    /* Process candidate key event before hot keys,
+     * or lookup table can't receive important events.
+     * For example, if Esc key is pressed, this key event should be used for
+     * closing lookup table, not for turning to latin mode. */
+    if (hangul->hanja_list != NULL) {
+        retval = ibus_hangul_engine_process_candidate_key_event (hangul,
+                     keyval, modifiers);
+        if (hangul->hanja_mode) {
+            if (retval)
+                return TRUE;
+        } else {
+            return TRUE;
+        }
+    }
+
     // If a hotkey has any modifiers, we ignore that modifier
     // keyval, or we cannot make the hanja key work.
     // Because when we get the modifier key alone, we commit the
@@ -1105,17 +1455,6 @@ ibus_hangul_engine_process_key_event (IBusEngine     *engine,
         return TRUE;
     }
 
-    if (hangul->hanja_list != NULL) {
-        retval = ibus_hangul_engine_process_candidate_key_event (hangul,
-                     keyval, modifiers);
-        if (hangul->hanja_mode) {
-            if (retval)
-                return TRUE;
-        } else {
-            return TRUE;
-        }
-    }
-
     // If we've got a key event with some modifiers, commit current
     // preedit string and ignore this key event.
     // So, if you want to add some key event handler, put it 
@@ -1128,6 +1467,10 @@ ibus_hangul_engine_process_key_event (IBusEngine     *engine,
         return FALSE;
     }
 
+    if (hangul->preedit_mode == PREEDIT_MODE_NONE) {
+        ibus_hangul_engine_check_caret_pos_sanity (hangul);
+    }
+
     if (keyval == IBUS_BackSpace) {
         retval = hangul_ic_backspace (hangul->context);
         if (!retval) {
@@ -1138,7 +1481,11 @@ ibus_hangul_engine_process_key_event (IBusEngine     *engine,
             }
         }
 
-        ibus_hangul_engine_update_preedit_text (hangul);
+        if (hangul->preedit_mode == PREEDIT_MODE_NONE) {
+            ibus_hangul_engine_process_commit_and_edit (hangul);
+        } else {
+            ibus_hangul_engine_update_preedit_text (hangul);
+        }
 
         if (hangul->hanja_mode) {
             if (ibus_hangul_engine_has_preedit (hangul)) {
@@ -1173,42 +1520,12 @@ ibus_hangul_engine_process_key_event (IBusEngine     *engine,
         }
         retval = hangul_ic_process (hangul->context, keyval);
 
-        str = hangul_ic_get_commit_string (hangul->context);
-        if (word_commit || hangul->hanja_mode) {
-            const ucschar* hic_preedit;
-
-            hic_preedit = hangul_ic_get_preedit_string (hangul->context);
-            if (hic_preedit != NULL && hic_preedit[0] != 0) {
-                ustring_append_ucs4 (hangul->preedit, str, -1);
-            } else {
-                IBusText *text;
-                const ucschar* preedit;
-
-                ustring_append_ucs4 (hangul->preedit, str, -1);
-                if (ustring_length (hangul->preedit) > 0) {
-                    /* clear preedit text before commit */
-                    ibus_hangul_engine_clear_preedit_text (hangul);
-
-                    preedit = ustring_begin (hangul->preedit);
-                    text = ibus_text_new_from_ucs4 ((gunichar*)preedit);
-                    ibus_engine_commit_text (engine, text);
-                }
-                ustring_clear (hangul->preedit);
-            }
+        if (hangul->preedit_mode == PREEDIT_MODE_NONE) {
+            ibus_hangul_engine_process_commit_and_edit (hangul);
         } else {
-            if (str != NULL && str[0] != 0) {
-                IBusText *text;
-
-                /* clear preedit text before commit */
-                ibus_hangul_engine_clear_preedit_text (hangul);
-
-                text = ibus_text_new_from_ucs4 (str);
-                ibus_engine_commit_text (engine, text);
-            }
+            ibus_hangul_engine_process_edit_and_commit (hangul);
         }
 
-        ibus_hangul_engine_update_preedit_text (hangul);
-
         if (hangul->hanja_mode) {
             ibus_hangul_engine_update_lookup_table (hangul);
         }
@@ -1262,7 +1579,7 @@ ibus_hangul_engine_process_key_event (IBusEngine     *engine,
      */
     if (use_event_forwarding) {
         if (!retval) {
-            ibus_engine_forward_key_event (engine, keyval, keycode, modifiers);
+            ibus_engine_forward_key_event (engine, orig_keyval, keycode, modifiers);
         }
 
         return TRUE;
@@ -1290,7 +1607,7 @@ ibus_hangul_engine_flush (IBusHangulEngine *hangul)
 	str = ustring_begin (hangul->preedit);
 	text = ibus_text_new_from_ucs4 (str);
 
-        g_debug("flush: %s", text->text);
+        g_debug ("flush:%u: %s", hangul->id, text->text);
 	ibus_engine_commit_text ((IBusEngine *) hangul, text);
 
 	ustring_clear(hangul->preedit);
@@ -1304,6 +1621,10 @@ ibus_hangul_engine_focus_in (IBusEngine *engine)
 {
     IBusHangulEngine *hangul = (IBusHangulEngine *) engine;
 
+    //g_debug ("focus_in: %u", hangul->id);
+
+    ibus_hangul_engine_update_preedit_mode (hangul);
+
     if (hangul->input_mode == INPUT_MODE_HANGUL) {
         ibus_property_set_state (hangul->prop_hangul_mode, PROP_STATE_CHECKED);
     } else {
@@ -1332,12 +1653,15 @@ ibus_hangul_engine_focus_out (IBusEngine *engine)
 {
     IBusHangulEngine *hangul = (IBusHangulEngine *) engine;
 
+    //g_debug ("focus_out: %u", hangul->id);
+
     if (hangul->hanja_list == NULL) {
 	// ibus-hangul uses
 	// ibus_engine_update_preedit_text_with_mode() function which makes
 	// the preedit string committed automatically when the focus is out.
 	// So we don't need to commit the preedit here.
 	hangul_ic_reset (hangul->context);
+        ustring_clear (hangul->preedit);
     } else {
         ibus_engine_hide_lookup_table (engine);
         ibus_engine_hide_auxiliary_text (engine);
@@ -1351,7 +1675,24 @@ ibus_hangul_engine_reset (IBusEngine *engine)
 {
     IBusHangulEngine *hangul = (IBusHangulEngine *) engine;
 
+    g_debug ("reset:%u", hangul->id);
+
+    if (hangul->preedit_mode == PREEDIT_MODE_NONE) {
+        hangul_ic_reset (hangul->context);
+        ustring_clear (hangul->preedit);
+    }
+
+    if (use_client_commit) {
+        // ibus-hangul uses
+        // ibus_engine_update_preedit_text_with_mode() function which makes
+        // the preedit string committed automatically when the reset is received
+        // So we don't need to commit the preedit here.
+        hangul_ic_reset (hangul->context);
+        ustring_clear (hangul->preedit);
+    }
+
     ibus_hangul_engine_flush (hangul);
+
     IBUS_ENGINE_CLASS (parent_class)->reset (engine);
 }
 
@@ -1360,16 +1701,32 @@ ibus_hangul_engine_enable (IBusEngine *engine)
 {
     IBUS_ENGINE_CLASS (parent_class)->enable (engine);
 
+    g_debug ("enable:%u", ((IBusHangulEngine*) engine)->id);
+
     ibus_engine_get_surrounding_text (engine, NULL, NULL, NULL);
 }
 
 static void
 ibus_hangul_engine_disable (IBusEngine *engine)
 {
+    g_debug ("disable:%u", ((IBusHangulEngine*) engine)->id);
+
     ibus_hangul_engine_focus_out (engine);
     IBUS_ENGINE_CLASS (parent_class)->disable (engine);
 }
 
+static void
+ibus_hangul_engine_set_capabilities (IBusEngine *engine, guint caps)
+{
+    IBusHangulEngine *hangul = (IBusHangulEngine *) engine;
+
+    hangul->caps = caps;
+
+    ibus_hangul_engine_update_preedit_mode (hangul);
+
+    g_debug ("set_capabilities:%u: %x", hangul->id, caps);
+}
+
 static void
 ibus_hangul_engine_page_up (IBusEngine *engine)
 {
@@ -1467,6 +1824,11 @@ ibus_hangul_engine_switch_input_mode (IBusHangulEngine *hangul)
         input_mode = INPUT_MODE_HANGUL;
     }
 
+    if (hangul->preedit_mode == PREEDIT_MODE_NONE) {
+        hangul_ic_reset (hangul->context);
+        ustring_clear (hangul->preedit);
+    }
+
     ibus_hangul_engine_set_input_mode (hangul, input_mode);
 }
 
@@ -1504,7 +1866,8 @@ ibus_hangul_engine_set_input_mode (IBusHangulEngine *hangul, int input_mode)
     prop = hangul->prop_hangul_mode;
 
     hangul->input_mode = input_mode;
-    g_debug("input_mode: %s", (input_mode == INPUT_MODE_HANGUL) ? "hangul" : "latin");
+    g_debug("input_mode:%u: %s", hangul->id,
+            (input_mode == INPUT_MODE_HANGUL) ? "hangul" : "latin");
 
     symbol = ibus_hangul_engine_get_input_mode_symbol (hangul, input_mode);
     ibus_property_set_symbol(prop, symbol);
@@ -1539,6 +1902,14 @@ ibus_hangul_engine_on_transition (HangulInputContext     *hic,
     return true;
 }
 
+static void
+print_changed_settings (const gchar *schema_id, const gchar *key, GVariant *value)
+{
+    gchar *variant_printable = g_variant_print (value, FALSE);
+    g_debug ("settings_changed: %s/%s: %s", schema_id, key, variant_printable);
+    g_free (variant_printable);
+}
+
 static void
 settings_changed (GSettings    *settings,
                   const gchar  *key,
@@ -1560,22 +1931,29 @@ settings_changed (GSettings    *settings,
             const gchar *str = g_variant_get_string(value, NULL);
             g_string_assign (hangul_keyboard, str);
             hangul_ic_select_keyboard (hangul->context, hangul_keyboard->str);
+            print_changed_settings (schema_id, key, value);
         } else if (strcmp (key, "hanja-keys") == 0) {
             const gchar* str = g_variant_get_string(value, NULL);
 	    hotkey_list_set_from_string(&hanja_keys, str);
+            print_changed_settings (schema_id, key, value);
         } else if (strcmp (key, "word-commit") == 0) {
             word_commit = g_variant_get_boolean (value);
+            print_changed_settings (schema_id, key, value);
         } else if (strcmp (key, "auto-reorder") == 0) {
             auto_reorder = g_variant_get_boolean (value);
+            print_changed_settings (schema_id, key, value);
         } else if (strcmp (key, "switch-keys") == 0) {
             const gchar* str = g_variant_get_string(value, NULL);
 	    hotkey_list_set_from_string(&switch_keys, str);
+            print_changed_settings (schema_id, key, value);
         } else if (strcmp (key, "on-keys") == 0) {
             const gchar* str = g_variant_get_string(value, NULL);
 	    hotkey_list_set_from_string(&on_keys, str);
+            print_changed_settings (schema_id, key, value);
         } else if (strcmp (key, "off-keys") == 0) {
             const gchar* str = g_variant_get_string(value, NULL);
 	    hotkey_list_set_from_string(&off_keys, str);
+            print_changed_settings (schema_id, key, value);
         } else if (strcmp (key, "initial-input-mode") == 0) {
             const gchar* str = g_variant_get_string (value, NULL);
             if (strcmp(str, "latin") == 0) {
@@ -1583,10 +1961,25 @@ settings_changed (GSettings    *settings,
             } else if (strcmp(str, "hangul") == 0) {
                 initial_input_mode = INPUT_MODE_HANGUL;
             }
+            print_changed_settings (schema_id, key, value);
+        } else if (strcmp (key, "use-event-forwarding") == 0) {
+            use_event_forwarding = g_variant_get_boolean (value);
+            print_changed_settings (schema_id, key, value);
+        } else if (strcmp (key, "preedit-mode") == 0) {
+            const gchar* str = g_variant_get_string (value, NULL);
+            if (strcmp(str, "none") == 0) {
+                global_preedit_mode = PREEDIT_MODE_NONE;
+            } else if (strcmp(str, "word") == 0) {
+                global_preedit_mode = PREEDIT_MODE_WORD;
+            } else {
+                global_preedit_mode = PREEDIT_MODE_SYLLABLE;
+            }
+            print_changed_settings (schema_id, key, value);
         }
     } else if (strcmp (schema_id, "org.freedesktop.ibus.panel") == 0) {
         if (strcmp (key, "lookup-table-orientation") == 0) {
             lookup_table_orientation = g_variant_get_int32(value);
+            print_changed_settings (schema_id, key, value);
         }
     }
     g_variant_unref (value);
